#include <stdio.h>
#include <stdlib.h>

/* Here we define another datatype useful for implementing queues, the
 * circular array.  The circular array is the same as a normal array,
 * except that we no longer assume that the data starts at the beginning
 * of the array; in this way, the queue can march through the allocated
 * memory freely as data is pushed and shifted.  Because the start point
 * is now unfixed, we must do index arithmetic modulo the capacity of
 * the array.
 */

struct queue
{
  unsigned int capacity; /* The allocated size of data. */
  unsigned int size;     /* The number of valid elements in the queue. */
  unsigned int start;    /* The index of the first valid element. */
  int *data;             /* The actual data, a dynamic array of ints. */
};

typedef struct queue queue;

int make_queue(queue *q, unsigned int capacity)
{
  if (q == NULL || capacity < 1)
    return -1;

  q->capacity = capacity;
  q->size = 0;
  q->start = 0;
  q->data = (int *) malloc(capacity * sizeof(int));
  if (q->data == NULL)
    return -1;

  return 0;
}

int push_queue(queue *q, int value)
{
  if (q == NULL || q->size >= q->capacity)
    return -1;

  q->data[(q->start + q->size) % q->capacity] = value;
  ++q->size;
  return 0;
}

int shift_queue(queue *q, int *value)
{
  if (q == NULL || q->size == 0)
    return -1;

  *value = q->data[q->start];
  q->start = (q->start + 1) % q->capacity;
  --q->size;
  return 0;
}

void debug_queue(queue *q)
{
  int i;

  if (q == NULL)
    return;
 
  for (i = 0; i < q->capacity; ++i)
  {
    if (i > 0 && i != (q->start + q->size) % q->capacity)
      printf(" ");
    if (i == q->start)
      printf("[");
    if (i == (q->start + q->size) % q->capacity)
      printf("] ");
    printf("%d", q->data[i]);
  }
  printf("\n");
}

void free_queue(queue *q)
{
  if (q == NULL)
    return;

  free(q->data);
}

int main(int argc, char **argv)
{
  queue q;
  int i;

  if (make_queue(&q, 10) < 0)
  {
    fprintf(stderr, "Unable to allocate memory for queue.\n");
    exit(1);
  }

  printf("The queue starts with %u elements.\n", q.size);
  debug_queue(&q);
  for (i = 0; i < 10; ++i)
  {
    push_queue(&q, i);
    debug_queue(&q);
  }

  while (q.size > 5)
  {
    shift_queue(&q, &i);
    debug_queue(&q);
    printf("Shifted off the value %d.\n", i);
  }

  for (i = 10; i < 15; ++i)
  {
    push_queue(&q, i);
    debug_queue(&q);
  }

  while (q.size > 0)
  {
    shift_queue(&q, &i);
    debug_queue(&q);
    printf("Shifted off the value %d.\n", i);
  }

  free_queue(&q);
  return 0;
}
